require 'optparse'

Dir[File.expand_path('../docker_rails_proxy/extends/*.rb', __FILE__)].map do |f|
  require f
end

module DockerRailsProxy
  NUMBER_REGEXP = /^\d+$/.freeze

  COMMANDS = Dir[File.expand_path('../docker_rails_proxy/commands/*.rb', __FILE__)].map do |f|
    File.basename(f, '.rb').parameterize
  end.freeze

  COMMANDS.each do |c|
    autoload c.classify.to_sym, "docker_rails_proxy/commands/#{c.underscore}"
  end

  autoload :InheritableAttributes, 'docker_rails_proxy/concerns/inheritable_attributes'
  autoload :Callbacks,             'docker_rails_proxy/concerns/callbacks'
  autoload :Rsync,                 'docker_rails_proxy/concerns/rsync'
  autoload :Logger,                'docker_rails_proxy/concerns/logger'

  autoload :Cli,     'docker_rails_proxy/cli'
  autoload :VERSION, 'docker_rails_proxy/version'

  class Base
    include InheritableAttributes
    include Callbacks
    include Rsync
    include Logger

    attr_accessor :additional_arguments_options
    attr_reader   :arguments, :additional_arguments

    after_initialize { self.additional_arguments_options = {} }
    after_initialize { additional_arguments_parser.parse!(additional_arguments) }
    after_initialize do
      DockerRailsProxy.log_level(level: additional_arguments_options[:debug])
    end

    def initialize(arguments:, additional_arguments:)
      @arguments            = arguments
      @additional_arguments = additional_arguments
    end

    class << self
      def build_path(path = nil)
        File.join(APP_PATH, block_given? ? yield : path)
      end

      def command
        name.sub('DockerRailsProxy::', ''.freeze).parameterize.sub('--', ' '.freeze)
      end

      def execute(options)
        system "#{build_path("bin/#{APP_NAME}")} #{command} #{options}"
      end

      def call(options)
        klass = _run_build_callbacks params: options

        klass.send :_run_before_initialize_callbacks

        resource = klass.new(**options)

        klass.send :_run_after_initialize_callbacks, resource: resource

        klass.send :_run_validation_callbacks, resource: resource

        klass.send :_run_before_process_callbacks, resource: resource

        result = resource.process

        klass.send :_run_after_process_callbacks, resource: resource
        result
      end
    end

    def process
      raise NotImplementedError, 'Subclasses must implement a process method'
    end

    private

    def build_path(*args, &block)
      self.class.build_path(*args, &block)
    end

    def print_options(values, message)
      puts message
      values.each_with_index { |v, i| puts "#{i}) #{v}" }
    end

    def get_option(values, default: nil, array: false)
      flush_stdin
      print ': '
      option = $stdin.gets.chomp

      return default if option.blank?

      if array
        options = option.split(',')
        return unless options.all? { |o| o =~ NUMBER_REGEXP }

        options.map { |o| values[o.to_i].split(',') }.flatten.compact.uniq.join(',')
      else
        values[option.to_i] if option =~ NUMBER_REGEXP
      end
    end

    def until_get_option(values, message, options = {})
      value = nil

      while value.nil? do
        print_options(values, message)
        value = get_option(values, **options)
      end

      value
    end

    def flush_stdin
      loop do
        Timeout::timeout(0.1) { $stdin.gets.chomp } rescue break
      end
    end

    def additional_arguments_parser
      @additional_arguments_parser ||= OptionParser.new do |opts|
        opts.banner = "Usage: bin/#{APP_NAME} #{self.class.name.demodulize.parameterize} [<arguments>] -- [<additional_arguments>]".bold

        opts.on('--debug', 'Debug mode') do |debug|
          additional_arguments_options[:debug] = ::Logger::DEBUG
        end

        yield opts if block_given?

        opts.on('-h', '--help', 'Display this screen') do
          puts opts
          exit
        end
      end
    end
  end

  class AwsCli < Base
    before_initialize do
      unless system 'type aws &> /dev/null'
        'aws is required, `brew install awscli`'
      end
    end
  end

  class Docker < Base
    class << self
      attr_accessor :vm_provisioner, :kubernetes_running
    end

    attr_accessor :app_container_id, :docker_options

    before_initialize do
      unless system 'type docker &> /dev/null'
        'Please install docker before continuing.'
      end
    end

    before_initialize do
      unless system 'docker ps &> /dev/null'
        %(
        Couldn't connect to Docker daemon you might need to run:
          docker-machine start default
          eval $(docker-machine env default) or eval $(minikube docker-env)
        )
      end
    end

    before_initialize do
      if ENV['DOCKER_CERT_PATH'].present?
        Docker.vm_provisioner = ENV['DOCKER_CERT_PATH'].include?('.minikube') ? :minikube : :docker_machine
      end

      Docker.vm_provisioner ||= :docker_for_mac
    end

    before_initialize do
      if system 'docker images | grep -c k8s &> /dev/null'
        Docker.kubernetes_running = true
      end
    end

    before_initialize do
      unless system 'type kubectl &> /dev/null'
        'kubectl is required, `brew install kubectl`'
      end if Docker.kubernetes_running
    end

    after_initialize { self.docker_options = {} }
    after_initialize :set_app_container_id

    private

    def docker_compose?
      !!Docker.kubernetes_running == false
    end

    def kubernetes?
      Docker.kubernetes_running
    end

    def set_app_container_id
      self.app_container_id = get_docker_container_id(APP_NAME)
    end

    def get_docker_container_id(app, container: nil)
      if docker_compose?
        %x(docker ps -q --filter "name=#{[app, container].compact.join('_')}$").strip
      elsif kubernetes?
        pod_name = %x(kubectl get pod -l "app=#{app}" 2> /dev/null | grep Running | awk '{print $1}').strip
        return '' if pod_name.empty?

        full_id  = %x(kubectl get pod #{pod_name} -o jsonpath='{ .status.containerStatuses[?(@.name=="#{container || app}")].containerID }').strip
        full_id.split('//').last
      end
    end

    def execute(command,
                tty: false,
                container_id: app_container_id,
                replace_process: false,
                **)

      docker_options['-ti'] = nil if tty

      command = [
        'docker exec',
        docker_options.map { |key, value| [key, value].compact.join(' ') },
        container_id,
        command
      ].join(' '.freeze)

      replace_process ? exec(command) : system(command)
    end

    def additional_arguments_parser
      super do |opts|
        opts.on('--docker-workdir WORKDIR', 'Working directory inside the container') do |workdir|
          docker_options['-w'] = workdir
        end

        opts.on('--docker-user USER', 'Username or UID (format: <name|uid>[:<group|gid>])') do |user|
          docker_options['-u'] = user
        end

        opts.on('--docker-tty', 'Allocate a pseudo-TTY') do |tty|
          docker_options['-t'] = nil
        end

        opts.on('--docker-interactive', 'Keep STDIN open even if not attached') do |interactive|
          docker_options['-i'] = nil
        end
      end
    end
  end

  class DockerMainApp < Docker
    after_initialize do
      if app_container_id.empty?
        %(
        Couldn't connect to #{APP_NAME}'s container you might need to run:
          bin/#{APP_NAME} compose up
        )
      end
    end

    # Makes sure the container is in sync with the APP_PATH
    before_process { sync source: APP_PATH, silent: true }
  end

  class SyncBack < DockerMainApp
    after_process { execute "chown -R nobody:nogroup ."  }
    after_process { sync source: APP_PATH, reverse: true }
  end
end
