module DockerRailsProxy
  module Rsync
    class << self
      attr_accessor :exclusions,
                    :exclusions_from_host_to_container,
                    :exclusions_from_container_to_host

      def included(base)
        base.extend(ClassMethods)
      end
    end

    self.exclusions = %w[
      tmp/backups
      tmp/cache
      tmp/letter_opener
      tmp/pids
      tmp/sockets
      tmp/sessions
      log
      .git*
    ]

    self.exclusions_from_host_to_container = []
    self.exclusions_from_container_to_host = []

    module ClassMethods
    end

    def sync(source:, reverse: false, silent: false)
      _rsync.sync source: source, reverse: reverse, silent: silent
    end

    private

    def _rsync_host
      @_rsync_host ||= begin
        ip   = ENV.fetch('DOCKER_HOST') { '127.0.0.1' }.to_s.sub('tcp://', '').sub(/:\d+$/, '')
        port = '10873'

        loop do
          break if (port = %x(kubectl get svc "#{APP_NAME}" -o jsonpath='{.spec.ports[?(@.name=="rsync")].nodePort}' 2> /dev/null).strip).present?
          sleep 5
          logger.info 'Waiting for rsync port ...'
        end if Docker.kubernetes_running

        "rsync://#{ip}:#{port}"
      end
    end

    def _rsync
      @_rsync ||= Sync.new(rsync_host: _rsync_host)
    end

    class Sync
      attr_reader :rsync_host

      def initialize(rsync_host:)
        @rsync_host = rsync_host
      end

      def self.call(rsync_host:, **options)
        new(rsync_host: rsync_host).sync(options)
      end

      def sync(options)
        return if source_in_exclusions?(**options)

        source, target, volume = normalize_options(**options)

        result = send(
          "sync_#{volume}", source: source, target: target, reverse: options[:reverse]
        )

        if result && options[:silent].eql?(false)
          DockerRailsProxy.logger.info "#{source}   =======>   #{target}"
        end

        result
      end

    private

      def normalize_options(source:, reverse: false, **)
        values = if source.include? APP_PATH
                   ["#{APP_PATH}/", 'app']
                 elsif GEMS_PATH.present? and source.include?(GEMS_PATH)
                   ["#{GEMS_PATH}/", 'gems']
                 else
                   $stderr.puts "There is no rsync volume related with this path: #{source}"
                   exit 1
                 end

        paths = [values.first, "#{rsync_host}/#{values.last}"]

        [paths.send(reverse ? :reverse : :to_a), values.last].flatten
      end

      def sync_app(source:, target:, reverse:)
        exclusions = all_exclusions(reverse: reverse)

        system <<-EOS
        rsync -avqP --no-owner --no-group \
        #{exclusions.map{ |e| "--exclude '#{e}'" }.join(' '.freeze)} \
          --force \
          --delete \
          #{source} #{target}
        EOS
      end

      def sync_gems(source:, target:, **)
        system <<-EOS
        rsync -avqP --no-owner --no-group \
          --exclude '.git*' \
          --force \
          --delete \
          #{source} #{target}
        EOS
      end

      def all_exclusions(reverse:)
        one_way_exclusions = if reverse
                               Rsync.exclusions_from_container_to_host
                              else
                                Rsync.exclusions_from_host_to_container
                              end

        Rsync.exclusions + one_way_exclusions
      end

      def source_in_exclusions?(source:, reverse:, **)
        !!all_exclusions(reverse: reverse).find do |e|
          source.include?("#{APP_PATH}/#{e}")
        end
      end
    end
  end
end
